/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * 4bpp VGA Planar Video Driver for MicroWindows
 *
 * Inspired from Ben Pfaff's BOGL <pfaffben@debian.org>
 */
/*#define NDEBUG*/
#include <assert.h>
#include <sys/io.h>
#include "../device.h"
#include "fb.h"

/* Program the Set/Reset Register for drawing in color COLOR for write
   mode 0. */
static inline void 
set_color (int c)
{
  outb (0, 0x3ce);
  outb (c, 0x3cf);
}

/* Set the Enable Set/Reset Register. */
static inline void
set_enable_sr (int mask)
{
  outb (1, 0x3ce);
  outb (mask, 0x3cf);
}

/* Select the Bit Mask Register on the Graphics Controller. */
static inline void
select_mask (void)
{
  outb (8, 0x3ce);
}

/* Program the Bit Mask Register to affect only the pixels selected in
   MASK.  The Bit Mask Register must already have been selected with
   select_mask (). */
static inline void
set_mask (int mask)
{
  outb (mask, 0x3cf);
}

/* Set the Data Rotate Register.  Bits 0-2 are rotate count, bits 3-4
   are logical operation (0=NOP, 1=AND, 2=OR, 3=XOR). */
static inline void
set_op (int op)
{
  outb (3, 0x3ce);
  outb (op, 0x3cf);
}

/* Set the Memory Plane Write Enable register. */
static inline void
set_write_planes (int mask)
{
  outb (2, 0x3c4);
  outb (mask, 0x3c5);
}

/* Set the Read Map Select register. */
static inline void
set_read_plane (int plane)
{
  outb (4, 0x3ce);
  outb (plane, 0x3cf);
}

/* Set the Graphics Mode Register.  The write mode is in bits 0-1, the
   read mode is in bit 3. */
static inline void
set_mode (int mode)
{
  outb (5, 0x3ce);
  outb (mode, 0x3cf);
}

/* Read-modify-write the specified memory byte. */
static inline void
rmw (volatile char *p)
{
  *p |= 1;
}

/* Set pixel at x, y, to pixelval c*/
void
vgaplan4_drawpixel(ADDR8 addr, COORD x, COORD y, PIXELVAL c)
{
    assert (x >= 0 && x < XRES);
    assert (y >= 0 && y < YRES);
    assert (c >= 0 && c < NCOLORS);
  
	++drawing;
    set_color (c);
    select_mask ();
    set_mask (0x80 >> (x % 8));
    rmw (addr + x / 8 + y * LINELEN);
    --drawing;
}

/* Return 4-bit pixel at x, y*/
PIXELVAL
vgaplan4_readpixel(ADDR8 addr, COORD x, COORD y)
{
	volatile char *	src;
	int		plane;
	int		c = 0;
	
	assert (x >= 0 && x < XRES);
	assert (y >= 0 && y < YRES);
  
	++drawing;
	src = addr + x / 8 + y * LINELEN;
	for(plane=0; plane<4; ++plane) {
		set_read_plane(plane);
		if(*src & (0x80 >> (x % 8)))
			c |= 1 << plane;
	}
	--drawing;
	return c;
}


/* Draw horizontal line from x1,y to x2,y not including final point*/
void
vgaplan4_drawhorzline(ADDR8 addr, COORD x1, COORD x2, COORD y, PIXELVAL c)
{
	volatile ADDR8	dst;
	volatile ADDR8	last;

	assert (x1 >= 0 && x1 < XRES);
	assert (x2 >= 0 && x2 <= XRES);
	assert (x2 >= x1);
	assert (y >= 0 && y < YRES);
	assert (c >= 0 && c < NCOLORS);

	if(x1 == x2)
		return;

	++drawing;
	set_color (c);
	dst = addr + x1 / 8 + y * LINELEN;
	select_mask ();
	if (x1 / 8 == x2 / 8) {
		set_mask ((0xff >> (x1 % 8)) & (0xff << (7 - x2 % 8)));
		rmw (dst);
	} else {
		set_mask (0xff >> (x1 % 8));
		rmw (dst++);

		set_mask (0xff);
		last = addr + x2 / 8 + y * LINELEN;
		while (dst < last)
			*dst++ = 1;
      
		set_mask (0xff << (7 - x2 % 8));
		rmw (dst);
	}
	--drawing;
}

/* Draw a vertical line from x,y1 to x,y2 not including final point*/
void
vgaplan4_drawvertline(ADDR8 addr, COORD x, COORD y1, COORD y2, PIXELVAL c)
{
	volatile ADDR8	dst;
	volatile ADDR8	last;

	y2--;	// FIXME

	assert (x >= 0 && x < XRES);
	assert (y1 >= 0 && y1 < YRES);
	assert (y2 >= 0 && y2 <= YRES);
	assert (y2 >= y1);
	assert (c >= 0 && c < NCOLORS);

	++drawing;
	set_color (c);
	select_mask ();
	set_mask (0x80 >> (x % 8));
	dst = addr + x / 8 + y1 * LINELEN;
	last = addr + x / 8 + y2 * LINELEN;
	while (dst <= last) {
		rmw (dst);
		dst += LINELEN;
	}
	--drawing;
}

void
vgaplan4_blit(ADDR8 dstaddr, COORD destx, COORD desty, COORD w, COORD h,
	ADDR8 srcaddr, COORD srcx, COORD srcy)
{
	// FIXME: no blit for vgaplan4!
}

/* Init VGA controller, return framebuffer mmap size*/
int
vgaplan4_init(void)
{
	if(ioperm (0x3c0, 0x20, 1) == -1)
		return failmsg("can't get IO permissions: %m");

	/* Set up some default values for the VGA Graphics Registers. */
	set_enable_sr (0xf);
	set_op (0);
	set_mode (0);

	/* return framebuffer mmap size*/
	return 0x10000;
}

FBENTRY fbvgaplan4 = {
	(void *)vgaplan4_init,
	(void *)vgaplan4_drawpixel,
	(void *)vgaplan4_readpixel,
	(void *)vgaplan4_drawhorzline,
	(void *)vgaplan4_drawvertline,
	(void *)vgaplan4_blit
};
